#include "types.h"
#include "loongarch.h"
#include "defs.h"
#include "param.h"
#include "fs.h"
#include "spinlock.h"
#include "sleeplock.h"
#include "file.h"
#include "stat.h"
#include "proc.h"
#include "mmap.h"
#include "memlayout.h"

static uint64 lookup_fixed_segment(uint64 start, uint64 end) {
    struct proc *p = myproc();
    struct vma_struct *vma;

    // 检查请求的范围是否在有效的虚拟地址范围内。
    if (end > MAXVA || end < start) {
        return -1;
    }

    // 遍历当前进程的虚拟内存区域列表，检查指定的范围是否已经被映射。
    for (vma = p->vma_list; vma != 0; vma = vma->next) {
        // 如果找到了与指定范围重叠的 VMA，返回错误。
        if (!(end <= vma->start || start >= vma->end)) {
            return -1;
        }
    }

    // 没有找到与指定范围重叠的 VMA，所以可以在这个范围内进行映射。
    return start;
}

static uint64 lookup_segment(uint64 sz) {
    struct proc *p = myproc();
    struct vma_struct *vma;

    // 从MMAPBASE开始查找。
    uint64 base = MMAPBASE;

    while (base + sz < USTACK_BASE-PGSIZE) {
        // 查找是否与任何现有的VMA重叠。
        for (vma = p->vma_list; vma != 0; vma = vma->next) {
            if ((base >= vma->start && base < vma->end) ||
                (base + sz > vma->start && base + sz <= vma->end) ||
                (base <= vma->start && base + sz >= vma->end)) {
                // 与一个现有的VMA重叠，尝试下一个地址。
                base = vma->end;
                break;
            }
        }

        if (vma == 0) {
            // 没有任何现有的VMA与新的地址范围重叠。
            return base;
        }
    }

    // 没有找到足够的空间。
    panic("lookup_segment: oversize");
    return (uint64)-1;
}

int mmap_file(uint64 start, uint64 len, int prot, int flags, struct file *f, int64 off)
{
    struct proc *p = myproc();
    struct vma_struct *vma;
    char * mem;

    // 分配一个新的虚拟内存区域（VMA）结构体，并初始化它。
    vma = (struct vma_struct *)kalloc();
    if (vma == 0) {
        return -1; // Out of memory.
    }

    vma->start = start;
    vma->end = start + len;
    vma->prot = prot;
    vma->flags = flags;
    vma->file = filedup(f); // 复制文件描述符，以防止它被关闭。
    vma->offset = off;

    // 将新的 VMA 添加到进程的 VMA 列表中。
    vma->next = p->vma_list;
    p->vma_list = vma;

    int tag = 0;
    if(prot & PROT_READ) {
        tag |= PTE_MAT;
    }
    if(prot & PROT_WRITE) {
        tag |= PTE_P | PTE_W;
    }

    // 读取文件内容并映射到内存中。
    for (uint64 a = start; a < start + len; a += PGSIZE) {
        mem = kalloc();
        if (mem == 0) {
            return -1; // Out of memory.
        }
        memset(mem, 0, PGSIZE);
        if (mappages(p->pagetable, a, PGSIZE, (uint64)mem, tag) < 0) {
			//TODO:撤销分配，否则内存泄露
			panic("mmap_file: failed");
            return -1; // mappages failed.
        }
		fileread(f, (uint64)mem, PGSIZE);
    }
    return 0;
}
int mmap_anonymous(uint64 start, uint64 len, int prot, int flags) {
    struct proc *p = myproc();
    struct vma_struct *vma;

    // Allocate a new VMA structure and initialize it.
    vma = (struct vma_struct *)kalloc();
    if (vma == 0) {
        return -1; // Out of memory.
    }

    vma->start = start;
    vma->end = start + len;
    vma->prot = prot;
    vma->flags = flags;
    vma->file = 0; // No associated file.
    vma->offset = 0;

    // Add the new VMA to the process's VMA list.
    vma->next = p->vma_list;
    p->vma_list = vma;

    // Allocate physical pages and map them into the new virtual memory region.
    for (uint64 a = start; a < start + len; a += PGSIZE) {
        char *mem = kalloc();
        if (mem == 0) {
            return -1; // Out of memory.
        }
        memset(mem, 0, PGSIZE);
        if (mappages(p->pagetable, a, PGSIZE, (uint64)mem, prot) < 0) {
			//TODO:撤销分配，否则内存泄露
            return -1; // mappages failed.
        }
    }

    return 0;
}


uint64 do_mmap(uint64 start, uint64 len, int prot, int flags, struct file *f, int64 off) {
    // 如果 f 不为空，需要进行一些额外的检查。
    if (f) {
        struct inode *ip = f->ip;

        // 检查 off 是否超过了文件的大小。
        if (off >= ip->size) {
            return -1;
        }

		//TODO:检查f是否为文件夹

        // 检查文件的权限是否与请求的保护级别相符。
        if ((f->readable ^ (prot & PROT_READ)) || (f->writable ^ ((prot & PROT_WRITE) >> 1))) {
            return -2;
        }
    }

    // 计算需要映射的页面数量。
    uint64 sz = PGROUNDUP(len);

    // 如果 MAP_FIXED 标志被设置，那么就必须在 start 指定的位置进行映射。
    // 否则，系统可以选择一个合适的位置进行映射。
    uint64 va;
    if (flags & MAP_FIXED) {
        va = lookup_fixed_segment(start, start + sz);
    } else {
        va = lookup_segment(sz);
    }

    // 如果虚拟地址查找失败，返回错误。
    if (va < 0) {
        return va;
    }

    // 根据是否指定了 MAP_ANONYMOUS 标志，调用合适的映射函数。
    int ret = 0;
    if (f) {
        ret = mmap_file(va, len, prot, flags, f, off);
    } else {
        ret = mmap_anonymous(va, len, prot, flags);
    }

    // 如果映射失败，返回错误。
    if (ret < 0) {
        return ret;
    }

    // 返回映射区域的起始地址。
    return va;
}