
//
// File-system system calls.
// Mostly argument checking, since we don't trust
// user code, and calls into file.c and fs.c.
//

#include "types.h"
#include "loongarch.h"
#include "defs.h"
#include "param.h"
#include "stat.h"
#include "spinlock.h"
#include "proc.h"
#include "fs.h"
#include "sleeplock.h"
#include "file.h"
#include "fcntl.h"

#define NULL 0
#include "mmap.h"
// Fetch the nth word-sized system call argument as a file descriptor
// and return both the descriptor and the corresponding struct file.
static int argfd(int n, int *pfd, struct file **pf) {
    int fd;
    struct file *f;

    if (argint(n, &fd) < 0)
        return -1;
    if(fd == -100)*pfd = fd;
    if (fd < 0 || fd >= NOFILE)
        return -1;
    if ((f = myproc()->ofile[fd]) == 0) {
        if(pfd) {
            *pfd = fd;
        }
        return -2;
    }
    if (pfd)
        *pfd = fd;
    if (pf)
        *pf = f;
    return 0;
}

// Allocate a file descriptor for the given file.
// Takes over file reference from caller on success.
static int fdalloc(struct file *f) {
    int fd;
    struct proc *p = myproc();

    for (fd = 0; fd < NOFILE; fd++) {
        if (p->ofile[fd] == 0) {
            p->ofile[fd] = f;
            return fd;
        }
    }
    return -1;
}

uint64 sys_dup(void) {
    struct file *f;
    int fd;

    if (argfd(0, 0, &f) < 0)
        return -1;
    if ((fd = fdalloc(f)) < 0)
        return -1;
    filedup(f);
    return fd;
}

uint64
sys_dup3(void)
{
  struct file *s_f, *t_f;
  int t_fd;
  struct proc *p = myproc();

  if(argfd(0, 0, &s_f) < 0)
    return -1;
  
  switch(argfd(1, &t_fd, &t_f)){
    case -1: return -1;
    case  0: 
        if(t_f == s_f) return -1;
        fileclose(t_f);
    default: p->ofile[t_fd] = filedup(s_f); // ofile 进程打开文件表
  }

  return t_fd;
}


uint64 sys_read(void) {
    struct file *f;
    int n;
    uint64 p;

    if (argfd(0, 0, &f) < 0 || argint(2, &n) < 0 || argaddr(1, &p) < 0)
        return -1;
    return fileread(f, p, n);
}

uint64 sys_write(void) {
    struct file *f;
    int n;
    uint64 p;

    if (argfd(0, 0, &f) < 0 || argint(2, &n) < 0 || argaddr(1, &p) < 0)
        return -1;

    return filewrite(f, p, n);
}

uint64 sys_close(void) {
    int fd;
    struct file *f;

    if (argfd(0, &fd, &f) < 0)
        return -1;
    myproc()->ofile[fd] = 0;
    fileclose(f);
    return 0;
}

uint64 sys_fstat(void) {
    struct file *f;
    uint64 st; // user pointer to struct stat

    if (argfd(0, 0, &f) < 0 || argaddr(1, &st) < 0)
        return -1;
    return filestat(f, st);
}
uint64 sys_ffstat(void) {
    struct file *f;
    uint64 st; // user pointer to struct stat

    if (argfd(0, 0, &f) < 0 || argaddr(1, &st) < 0)
        return -1;
    return origin_filestat(f, st);
}
// Create the path new as a link to the same inode as old.
uint64 sys_link(void) {
    char name[DIRSIZ], new[MAXPATH], old[MAXPATH];
    struct inode *dp, *ip;

    if (argstr(0, old, MAXPATH) < 0 || argstr(1, new, MAXPATH) < 0)
        return -1;

    begin_op();
    if ((ip = namei(old)) == 0) {
        end_op();
        return -1;
    }

    ilock(ip);
    if (ip->type == T_DIR) {
        iunlockput(ip);
        end_op();
        return -1;
    }

    ip->nlink++;
    iupdate(ip);
    iunlock(ip);

    if ((dp = nameiparent(new, name)) == 0)
        goto bad;
    ilock(dp);
    if (dp->dev != ip->dev || dirlink(dp, name, ip->inum) < 0) {
        iunlockput(dp);
        goto bad;
    }
    iunlockput(dp);
    iput(ip);

    end_op();

    return 0;

bad:
    ilock(ip);
    ip->nlink--;
    iupdate(ip);
    iunlockput(ip);
    end_op();
    return -1;
}

// Is the directory dp empty except for "." and ".." ?
static int isdirempty(struct inode *dp) {
    int off;
    struct dirent de;

    for (off = 2 * sizeof(de); off < dp->size; off += sizeof(de)) {
        if (readi(dp, 0, (uint64)&de, off, sizeof(de)) != sizeof(de))
            panic("isdirempty: readi");
        if (de.inum != 0)
            return 0;
    }
    return 1;
}
static int unlink(char* path,int fd){
struct inode *ip, *dp;
    struct dirent de;
    char name[DIRSIZ];
    uint off;
	if(*path == '.'&&*(path+1) != '.'){
	nameiparent(path, name);
	dp = myproc()->cwd;
	}
    else if ((dp = nameiparent(path, name)) == 0) {
        return -1;
    }
    ilock(dp);

    // Cannot unlink "." or "..".
    if (namecmp(name, ".") == 0 || namecmp(name, "..") == 0)
        goto bad;
  iunlockput(dp);
    if ((ip = dirlookup(dp, name, &off)) == 0){
     ilock(dp);
        goto bad;
        }
     ilock(dp);
    ilock(ip);
    if (ip->nlink < 1)
        panic("unlink: nlink < 1");
    if (ip->type == T_DIR && !isdirempty(ip)) {
        iunlockput(ip);
        goto bad;
    }
    memset(&de, 0, sizeof(de));
   if (writei(dp, 0, (uint64)&de, off, sizeof(de)) != sizeof(de))
        panic("unlink: writei");
    if (ip->type == T_DIR) {
        dp->nlink--;
        iupdate(dp);
    }
    iunlockput(dp);
    ip->nlink--;
    iupdate(ip);
    iunlockput(ip);
    return 0;

bad:
    iunlockput(dp);
    return -1;
}
uint64 sys_unlink(void) {
    struct inode *ip, *dp;
    struct dirent de;
    char name[DIRSIZ], path[MAXPATH];
    uint off;

    if (argstr(0, path, MAXPATH) < 0)
        return -1;

    begin_op();
    if ((dp = nameiparent(path, name)) == 0) {
        end_op();
        return -1;
    }

    ilock(dp);

    // Cannot unlink "." or "..".
    if (namecmp(name, ".") == 0 || namecmp(name, "..") == 0)
        goto bad;

    if ((ip = dirlookup(dp, name, &off)) == 0)
        goto bad;
    ilock(ip);

    if (ip->nlink < 1)
        panic("unlink: nlink < 1");
    if (ip->type == T_DIR && !isdirempty(ip)) {
        iunlockput(ip);
        goto bad;
    }

    memset(&de, 0, sizeof(de));
    if (writei(dp, 0, (uint64)&de, off, sizeof(de)) != sizeof(de))
        panic("unlink: writei");
    if (ip->type == T_DIR) {
        dp->nlink--;
        iupdate(dp);
    }
    iunlockput(dp);

    ip->nlink--;
    iupdate(ip);
    iunlockput(ip);

    end_op();

    return 0;

bad:
    iunlockput(dp);
    end_op();
    return -1;
}

uint64
sys_unlinkat(void)
{
	
	char path[MAXPATH];
	char h[MAXPATH];
	int dirfd, mode;
	struct file *f = NULL;
	struct inode *dp = NULL, *ip;
	begin_op();
	if (argfd(0, &dirfd, &f) < 0) {
		if (dirfd != AT_FDCWD) {//!= AT_FDCWD
			end_op();
			return -1;
		}
		dp = myproc()->cwd;
	} else {
		
		dp = f->ip;
		if (dp->type ==  T_DIR) {
			end_op();
			return -1;
		}
	}

	int len;
	if((len = argstr(1, path, MAXPATH)) <= 0 || argint(2, &mode) < 0){
		end_op();
		return -1;
		}
	char *s = path + len - 1;
	while (s >= path && *s == '/') {
		s--;
	}
	for(int i = 0;i < MAXPATH;i++){
	h[i] = path[i];
	}
	if (s >= path && *s == '.' && (s == path || *--s == '/')) {
		printf("sys_unlinkat", "illegel path %s\n", path);
		end_op();
		return -1;
	}
	
	if ((ip = ename( path,dirfd)) == NULL) {
		printf("sys_unlinkat", "can namei %s\n", path);
		end_op();
		return -1;
	}
	//int isdir = S_ISDIR(ip->mode);
	if (ip->type == T_DIR && mode != AT_REMOVEDIR) {
		iput(ip);
		printf("sys_unlinkat", "illegel mode 0x%x against 0x%x\n", mode, ip->type);
		end_op();
		return -1;
	} else if (ip->type != T_DIR && mode == AT_REMOVEDIR) {
		iput(ip);
		end_op();
		return -1;
	}
	ilock(ip);
	if (ip->type == T_DIR&& isdirempty(ip) != 1) {
		iunlockput(ip);
		printf("sys_unlinkat", "dir isn't empty\n");
		end_op();
		return -1;
	}
	getcwd(ip,path,128);
	int ret ;
	iunlockput(ip);
	if(h[0] == '.'&&h[1] == '/'){
	ret = unlink(h+2,dirfd);
 	}
	else{	
	ret = unlink(h,dirfd);
	}
	end_op();
	return ret;
}
static struct inode *create(char *path, short type, short major, short minor) {
    struct inode *ip, *dp;
    char name[DIRSIZ];

    if ((dp = nameiparent(path, name)) == 0)
        return 0;

    ilock(dp);

    if ((ip = dirlookup(dp, name, 0)) != 0) {

        iunlockput(dp);
        ilock(ip);
        if (type == T_FILE && (ip->type == T_FILE || ip->type == T_DEVICE))
            return ip;
        iunlockput(ip);
        return 0;
     }
    if ((ip = ialloc(dp->dev, type)) == 0)
        panic("create: ialloc");
    ilock(ip);
    ip->major = major;
    ip->minor = minor;
    ip->nlink = 1;
    iupdate(ip);
    if (type == T_DIR) { // Create . and .. entries.
        dp->nlink++;     // for ".."
        iupdate(dp);
        // No ip->nlink++ for ".": avoid cyclic ref count.
        if (dirlink(ip, ".", ip->inum) < 0 || dirlink(ip, "..", dp->inum) < 0)
            panic("create dots");
    }
    if (dirlink(dp, name, ip->inum) < 0)
        panic("create: dirlink");
    iunlockput(dp);

    return ip;
}
uint64 sys_open(void) {
    char path[MAXPATH];
    int fd, omode;
    struct file *f;
    struct inode *ip;
    int n;

    if ((n = argstr(0, path, MAXPATH)) < 0 || argint(1, &omode) < 0)
        return -1;

    begin_op();

    if (omode & O_CREATE) {
        ip = create(path, T_FILE, 0, 0);
        if (ip == 0) {
            end_op();
            return -1;
        }
    } else {
        if ((ip = namei(path)) == 0) {
            end_op();
            return -1;
        }
        ilock(ip);
        if (ip->type == T_DIR && omode != O_RDONLY) {
            iunlockput(ip);
            end_op();
            return -1;
        }
    }

    if (ip->type == T_DEVICE && (ip->major < 0 || ip->major >= NDEV)) {
        iunlockput(ip);
        end_op();
        return -1;
    }

    if ((f = filealloc()) == 0 || (fd = fdalloc(f)) < 0) {
        if (f)
            fileclose(f);
        iunlockput(ip);
        end_op();
        return -1;
    }

    if (ip->type == T_DEVICE) {
        f->type = FD_DEVICE;
        f->major = ip->major;
    } else {
        f->type = FD_INODE;
        f->off = 0;
    }
    f->ip = ip;
    f->readable = !(omode & O_WRONLY);
    f->writable = (omode & O_WRONLY) || (omode & O_RDWR);

    if ((omode & O_TRUNC) && ip->type == T_FILE) {
        itrunc(ip);
    }

    iunlock(ip);
    end_op();
	
    return fd;
}

static struct inode*
create2(char *path, short type, int mode, int dirfd,int major,int minor)
{
    struct inode *ip, *dp;
    char name[DIRSIZ + 1];
  if((dp = enameparent(path, name, dirfd)) == 0)
    return 0;
  ilock(dp);
    if ((ip = dirlookup(dp, name, 0)) != 0) {
        iunlockput(dp);
        ilock(ip);
        if (type == T_FILE && (ip->type == T_FILE || ip->type == T_DEVICE))
            return ip;
        iunlockput(ip);
        return 0;
     }
  if ((ip = ialloc(dp->dev, type)) == 0) {
    iunlock(dp);
    iput(dp);
    return 0;
  }
    ilock(ip);
    ip->major = major;
    ip->minor = minor;
    ip->nlink = 1;
    iupdate(ip);
  if ((type == T_DIR && !(ip->type == T_DIR)) ||
      (type == T_FILE && (ip->type == T_DIR))) {
    iunlock(dp);
    iput(ip);
    iput(dp);
    return 0;
  }
   if (type == T_DIR) { // Create . and .. entries.
        dp->nlink++;     // for ".."
        iupdate(dp);
        // No ip->nlink++ for ".": avoid cyclic ref count.
        if (dirlink(ip, ".", ip->inum) < 0 || dirlink(ip, "..", dp->inum) < 0)
            panic("create dots");
    }
    if (dirlink(dp, name, ip->inum) < 0)
        panic("create: dirlink");
    iunlockput(dp);

  return ip;
}

uint64 openat(int dirfd,char* path, int flags,int mode)
{
  struct file *f;
  struct inode *ep;
  int fd;
	
  if(flags & O_CREATE) {
    ep = create2(path, T_FILE, flags, dirfd,0,0);
    if(ep == 0) {
      return -1;
    } 
  } else {
    if((ep = ename(path, dirfd))== 0) {
      return -1;
    }
    ilock(ep);

  }
  if(flags == O_DIRECTORY){
  ep->type = T_DIR;
  }

  if((f = filealloc()) == 0 || (fd = fdalloc(f)) < 0){
 
    if (f) {
      fileclose(f);
    }
    iunlock(ep);
    iput(ep);
    return -1;
  }

  if(!(ep->type == T_DIR) && (flags & O_TRUNC)){
    itrunc(ep);
  }

  f->type = FD_INODE;
 
  f->off =   (flags & 0x004)  ? ep->size : 0;
  f->ip = ep;
  f->readable = !(flags & O_WRONLY);
  f->writable = (flags & O_WRONLY) || (flags & O_RDWR);
  iunlock(ep);

  return fd;
}



uint64
sys_openat(void)
{
  char path[MAXPATH];
  int fd, flags,mode;
  
  if(argint(0, &fd) < 0 || argstr(1, path, MAXPATH) < 0 || argint(2, &flags) < 0 || argint(3, &mode) < 0 )
    return -1;
     begin_op();

  if(fd < 0 && fd != -100) {
    printf("[sys_openat]Invalid file descriptor %d\n", fd);
    end_op();
    return -1;
  }

 // printf("[sys_openat] fd= %d\n", fd);


  int ret = openat( fd,path, flags,mode);

  //printf("[sys_openat] new fd = %d\n", ret);
   end_op();

  return ret;
}


uint64 sys_mkdir(void) {
    char path[MAXPATH];
    struct inode *ip;

    begin_op();
    if (argstr(0, path, MAXPATH) < 0 || (ip = create(path, T_DIR, 0, 0)) == 0) {
        end_op();
        return -1;
    }
    iunlockput(ip);
    end_op();
    return 0;
}
uint64
mkdirat(int dirfd, char* path, int mode)
{
  struct inode *ep;

  if((ep = create2(path, T_DIR, mode, dirfd,0,0)) == 0){
    return -1;
  }
  iunlock(ep);
  iput(ep);
  return 0;
}

uint64
sys_mkdirat(void)
{
  char path[MAXPATH];
  int fd, mode;

  if(argint(0, &fd) < 0 || argstr(1, path, MAXPATH) < 0 || argint(2, &mode) < 0){
    return -1;
  }
  begin_op();
  if(fd < 0 && fd != AT_FDCWD) {
    printf("[sys_mkdirat] Invalid file descriptor %d\n", fd);
    end_op();
    return -1;
  }


  int t = mkdirat(fd, path, mode);
    end_op();
    return t;
}
uint64 sys_mknod(void) {
    struct inode *ip;
    char path[MAXPATH];
    int major, minor;

    begin_op();
    if ((argstr(0, path, MAXPATH)) < 0 || argint(1, &major) < 0 ||
        argint(2, &minor) < 0 ||
        (ip = create(path, T_DEVICE, major, minor)) == 0) {
        end_op();
        return -1;
    }
    iunlockput(ip);
    end_op();
    return 0;
}

uint64 sys_chdir(void) {
    char path[MAXPATH];
    struct inode *ip;
    struct proc *p = myproc();

    begin_op();
    if (argstr(0, path, MAXPATH) < 0 || (ip = namei(path)) == 0) {
        end_op();
        return -1;
    }
    ilock(ip);
    if (ip->type != T_DIR) {
        iunlockput(ip);
        end_op();
        return -1;
    }
    iunlock(ip);
    iput(p->cwd);
    end_op();
    p->cwd = ip;
    return 0;
}

uint64 sys_exec(void) {
    char path[MAXPATH], *argv[MAXARG];
    int i;
    uint64 uargv, uarg;

    if (argstr(0, path, MAXPATH) < 0 || argaddr(1, &uargv) < 0) {
        return -1;
    }
    memset(argv, 0, sizeof(argv));
    for (i = 0;; i++) {
        if (i >= NELEM(argv)) {
            goto bad;
        }
        if (fetchaddr(uargv + sizeof(uint64) * i, (uint64 *)&uarg) < 0) {
            goto bad;
        }
        if (uarg == 0) {
            argv[i] = 0;
            break;
        }
        argv[i] = kalloc();
        if (argv[i] == 0)
            goto bad;
        if (fetchstr(uarg, argv[i], PGSIZE) < 0)
            goto bad;
    }

    int ret = exec(path, argv);

    for (i = 0; i < NELEM(argv) && argv[i] != 0; i++)
        kfree(argv[i]);

    return ret;

bad:
    for (i = 0; i < NELEM(argv) && argv[i] != 0; i++)
        kfree(argv[i]);
    return -1;
}
uint64
sys_execve(void)
{
  char path[MAXPATH], *argv[MAXARG],*env[MAXARG];
  int i;
  uint64 uargv, uarg, uenvp;
  if(argstr(0, path, MAXPATH) < 0 || argaddr(1, &uargv) < 0 || argaddr(1, &uenvp) < 0){
    return -1;
  }
  memset(argv, 0, sizeof(argv));
  memset(env,  0, sizeof(env));
  for(i=0;; i++){
    if(i >= NELEM(argv)){
      goto bad;
    }
    if(fetchaddr(uargv+sizeof(uint64)*i, (uint64*)&uarg) < 0){
      goto bad;
    }
    if(uarg == 0){
      argv[i] = 0;
      break;
    }
    argv[i] = kalloc();
    if(argv[i] == 0)
      goto bad;
    if(fetchstr(uarg, argv[i], PGSIZE) < 0)
      goto bad;
  }
  int ret = execve(path, argv, env);

  for(i = 0; i < NELEM(argv) && argv[i] != 0; i++)
    kfree(argv[i]);
  
  for(i = 0; i < NELEM(env) && env[i] != 0; i++)
    kfree(env[i]);

  return ret;

 bad:
  for(i = 0; i < NELEM(argv) && argv[i] != 0; i++)
    kfree(argv[i]);
  for(i = 0; i < NELEM(env) && env[i] != 0; i++)
    kfree(env[i]);
  return -1;

}

uint64 sys_pipe(void) {
    uint64 fdarray; // user pointer to array of two integers
    struct file *rf, *wf;
    int fd0, fd1;
    struct proc *p = myproc();

    if (argaddr(0, &fdarray) < 0)
        return -1;
    if (pipealloc(&rf, &wf) < 0)
        return -1;
    fd0 = -1;
    if ((fd0 = fdalloc(rf)) < 0 || (fd1 = fdalloc(wf)) < 0) {
        if (fd0 >= 0)
            p->ofile[fd0] = 0;
        fileclose(rf);
        fileclose(wf);
        return -1;
    }
    if (copyout(p->pagetable, fdarray, (char *)&fd0, sizeof(fd0)) < 0 ||
        copyout(p->pagetable, fdarray + sizeof(fd0), (char *)&fd1,
                sizeof(fd1)) < 0) {
        p->ofile[fd0] = 0;
        p->ofile[fd1] = 0;
        fileclose(rf);
        fileclose(wf);
        return -1;
    }
    return 0;
}
uint64
sys_getdents(void)
{
	struct file *f;
	uint64 p, len;

	if(argfd(0, 0, &f) < 0 || argaddr(1, &p) < 0 || argint(2, (int*)&len) < 0)
		return -1;
	return filereaddir(f, p, len);
}

uint64 sys_mmap(void)
 {
   // to gain the arguments , file and proc
   uint64 addr;
   int len , prot , flags , fd , off;
   if( argaddr( 0 , &addr ) < 0 || argint( 1 , &len ) < 0 || argint( 2 , &prot ) < 0 || argint( 3 , &flags ) < 0 || argint( 4 , &fd ) < 0 || argint( 5 , &off ) < 0 )
     return -1;
   struct proc* p = myproc();
   struct file* f = p->ofile[fd];
   
   // to ensure the prot
   if( ( flags == MAP_SHARED && f->writable == 0 && (prot&PROT_WRITE)) )
     return -1;
  
   // to find a empty vma and init it
   int idx;
   for( idx = 0 ; idx < VMA_MAX ; idx++ )
     if( p->vma[idx].valid == 0 )
       break;
   if( idx == VMA_MAX )
     panic("no empty vma field");
   
   struct VMA* vp = &p->vma[idx];
//    printf("===mmap===\n");
//    printf("p->vma[%d] success\n", idx);
   
   vp->valid = 1;
   vp->len = len;
   vp->flags = flags;
   vp->off = off;
   vp->prot = prot;
   vp->f = f;
   filedup( f );  //increase the ref of the file
   vp->addr = (p->maxaddr-=len);  // asign a useable virtual address to the vma field , and maintain the maxaddr
//    printf("p->vma[%d]->addr: %p", idx, p->vma[idx].addr);
   return vp->addr;
 }
uint64 sys_munmap(void)
 {
   // to gain the arguments and proc
   uint64 addr;
   int len;
   if( argaddr( 0 , &addr ) < 0 || argint( 1 , &len ) < 0 )
     return -1;
   struct proc* p = myproc();
   //to find the target vma
   struct VMA* vp = 0;
//    printf("===munmap===\n");
   for( int i=0 ; i<VMA_MAX ; i++ ) {
        // printf("p->vma[%d].addr: %p\n",i, p->vma[i].addr);
        // printf("p->vma[%d].addr+len: %p\n",i, p->vma[i].addr+p->vma[i].len);
        // printf("p->vma[%d].valid: %d\n",i, p->vma[i].valid);
     if( p->vma[i].addr <= addr && addr < p->vma[i].addr + p->vma[i].len && p->vma[i].valid == 1 )
     {
       vp = &p->vma[i];
       break;
     }
   }
   if( vp == 0 )
     panic("munmap no such vma");  
   // if the page has been mapped 
   if( walkaddr( p->pagetable , addr ) != 0)
   {
     //write back if necessary 
     if( vp->flags == MAP_SHARED )
       filewriteoff( vp->f , addr , len , addr-vp->addr ); // this function is new

     addr = PGROUNDDOWN(addr);
     uvmunmap( p->pagetable , addr , len/PGSIZE , 1 );   //unmap
     return 0;
   }
   // maintain the ref of file and the valid of the vma
   if( 0 == (vp->mapcnt -= len) )
   {
     fileclose( vp->f );
     vp->valid = 0;
   }
   return 0;
 }