#include "types.h"
#include "loongarch.h"
#include "defs.h"
#include "date.h"
#include "param.h"
#include "memlayout.h"
#include "spinlock.h"
#include "proc.h"
#include "sysinfo.h"
#include "struct_timespec.h"
#include "time.h"



uint64
sys_exit(void)
{
  int n;
  if(argint(0, &n) < 0)
    return -1;
  exit(n);
  return 0;  // not reached
}

uint64
sys_getpid(void)
{
  return myproc()->pid;
}
uint64
sys_getppid(void)
{
  if (myproc()->parent != 0) {
    return myproc()->parent->pid;
  }
  return -1;
}

uint64
sys_fork(void)
{
  return fork();
}

uint64
sys_wait(void)
{
  uint64 p;
  if(argaddr(0, &p) < 0)
    return -1;
  return wait(p);
}

uint64
sys_sbrk(void)
{
  int addr;
  int n;

  if(argint(0, &n) < 0)
    return -1;
  addr = myproc()->sz;
  if(growproc(n) < 0)
    return -1;
  return addr;
}

uint64
sys_brk(void)
{
  int addr;
  int new_addr;

  if(argint(0, &new_addr) < 0)
    return -1;

  addr = myproc()->sz;

  if(!new_addr) return addr;

  if(growproc(new_addr-addr) < 0)
    return -1;
  return 0;
}

uint64
sys_sleep(void)
{
  int n;
  uint ticks0;

  if(argint(0, &n) < 0)
    return -1;
  acquire(&tickslock);
  ticks0 = ticks;
  while(ticks - ticks0 < n){
    if(myproc()->killed){
      release(&tickslock);
      return -1;
    }
    sleep(&ticks, &tickslock);
  }
  release(&tickslock);
  return 0;
}

uint64
sys_kill(void)
{
  int pid;

  if(argint(0, &pid) < 0)
    return -1;
  return kill(pid);
}

// return how many clock tick interrupts have occurred
// since start.
uint64
sys_uptime(void)
{
  uint xticks;

  acquire(&tickslock);
  xticks = ticks;
  release(&tickslock);
  return xticks;
}

uint64
sys_clone(void){
  int flag;
  uint64 stack;
  if(argint(0, &flag) < 0) {
    return -1;
  }
  if(argaddr(1, &stack) < 0){
    return -1;
  }
  return clone(flag, stack);
}

uint64
sys_uname(void)
{
    uint64 addr;
  if (argaddr(0, &addr) < 0) {
    return -1;
  }

  struct utsname info;
  memmove(info.sysname, "USTC_La_XV6OS", strlen("USTC_La_XV6OS") + 1);
  memmove(info.nodename,"JustDo", strlen("JustDo") + 1);
  memmove(info.release, "0.0.0.1", strlen("0.0.0.1") + 1);
  memmove(info.version, "0.0.0.1", strlen("0.0.0.1") + 1);
  memmove(info.machine, "2k-500", strlen("2k-500") + 1);
  memmove(info.domainname, "domain", strlen("domain") + 1);

  if (copyout(myproc()->pagetable,addr, (char *)&info, sizeof(info)) < 0) {
    return -1;
  }
  return 0;
}

typedef struct timeval TimeVal;

uint64
sys_gettimeofday(void)
{
	TimeVal tval;
	uint64 tmp_ticks = r_time();
	tval.sec = tmp_ticks / CLK_FREQ;
  tval.usec = (tmp_ticks % CLK_FREQ) * 1000000 / CLK_FREQ;

	uint64 ptval;
	if (argaddr(0, &ptval) < 0) {
		return -1;
	}

	if (ptval && copyout((myproc()->pagetable), ptval, (char*)&tval, sizeof(tval)) < 0) {
		return -1;
	}
	return 0;
}

uint64
sys_times(void){
  struct proc* pproc = myproc();
  uint64 ptr;
  struct tms temp_time;
  if(argaddr(0, &ptr) < 0){return -1;}
  temp_time.tms_cutime = pproc->cutime;
  temp_time.tms_cstime = pproc->cstime;
  temp_time.tms_utime = pproc->user_time;// 这两个变量在系统trap中计数
  temp_time.tms_stime = pproc->kernel_time;

	if (ptr && copyout((myproc()->pagetable), ptr, (char*)&temp_time, sizeof(temp_time)) < 0) {
		return -1;
	}
  return 0;
}

uint64 sys_wait4(void)
{
  int pid, options;
  uint64 status;
  if(argint(0, &pid) < 0){
	  return -1;
  }
  if(argaddr(1, &status) < 0){
	  return -1;
  }
  if(argint(2, &options) < 0){
	  return -1;
  }
  return waitpid(pid, status, options);
}


uint64 sys_sched_yield(void)
{
  yield();
  return 0;
}

uint64 sys_nanosleep(void) {
	uint64 ureq, urem;
	struct timespec req, rem;

  // int now_lk = mycpu()->noff;
  // printf("before_lk: %d\n", now_lk);

	argaddr(0, &ureq);
	argaddr(1, &urem);
	struct proc *p = myproc();
	if (copyin((p->pagetable), (char*)&req, ureq, sizeof(struct timespec)) < 0) 
		return -1;
	if (req.tv_nsec > 999999999L) {
		return -1;
  }

  uint64 time = r_time() * 1000000 / CLK_FREQ;
	int ret = 0;
	uint64 expire;
  acquire(&tickslock);
  // 将时间单位调整到100ms
	expire = time + (req.tv_sec * 1000000 + req.tv_nsec );

  while (time < expire) 
  {
    // printf("%d %d %d\n", ticks, ticks_limit, n);
    if(p->killed){
      release(&tickslock);
      ret = -1;
      break;
    }
    sleep(&ticks, &tickslock);
    time = r_time() * 1000000 / CLK_FREQ;
  }

	if (ret < 0 && urem) {
    time = r_time() * 1000000 / CLK_FREQ;
		if (time < expire) {
      uint64 n = expire - time;
      rem.tv_sec = n / 10;
			rem.tv_nsec = n - rem.tv_sec * 100000;
			copyout((p->pagetable), urem, (char *)&rem, sizeof(struct timespec));
		}
	}
  release(&tickslock);
  // now_lk = mycpu()->noff;
  // printf("after_lk: %d\n", now_lk);
	return ret;
}


uint64
sys_getcwd(void){
  uint64 addr;
  int len;
  if(argaddr(0, &addr) < 0){
    return -1;
  }
  if(argint(1, &len) < 0){
    return -1;
  }
  char buf[MAXPATH];
  int max = MAXPATH < len ? MAXPATH:len;
  if(getcwd(myproc()->cwd,buf,max) < 0){
    return -1;
  }
  if(copyout(myproc()->pagetable,addr,buf,max)<0){
    return -1;
  }
  return addr;
}