#include "types.h"
#include "loongarch.h"
#include "defs.h"
#include "param.h"
#include "fs.h"
#include "spinlock.h"
#include "sleeplock.h"
#include "file.h"
#include "stat.h"
#include "proc.h"
#include "mmap.h"
#include "memlayout.h"

struct spinlock tickslock;
uint ticks;

// in kernelvec.S, calls kerneltrap().
void kernelvec();
void uservec();
void handle_excep();
void handle_tlbr();
void handle_merr();
void userret(uint64, uint64);
extern int devintr();

void
trapinit(void)
{
  initlock(&tickslock, "time");
  uint32 ecfg = ( 0U << CSR_ECFG_VS_SHIFT ) | HWI_VEC | TI_VEC;
  uint64 tcfg = 0x1000000UL | CSR_TCFG_EN | CSR_TCFG_PER;
  w_csr_ecfg(ecfg);
  w_csr_tcfg(tcfg);
  w_csr_eentry((uint64)handle_excep);
  w_csr_tlbrentry((uint64)handle_tlbr); //TODO
  w_csr_merrentry((uint64)handle_merr); //TODO
  intr_on();
}
//
// handle an interrupt, exception, or system call from user space.
// called from trampoline.S
//
void
usertrap(void)
{
  int which_dev = 0;

  if((r_csr_prmd() & PRMD_PPLV) == 0)
    panic("usertrap: not from user mode");

  struct proc *p = myproc();
  
  // save user program counter.
  p->trapframe->era = r_csr_era();
  
  if(((r_csr_estat()& CSR_ESTAT_ECODE) >> 16) == 0xb){
    // system call

    if(p->killed)
      exit(-1);

    // sepc points to the ecall instruction,
    // but we want to return to the next instruction.
    p->trapframe->era += 4;

    // an interrupt will change sstatus &c registers,
    // so don't enable until done with those registers.
    intr_on();

    syscall();
  } else if((which_dev = devintr()) != 0){
    // ok
  } 
  else if((r_csr_estat() >> 16) == PIL || (r_csr_estat() >> 16) == PIS) {
    // printf("=== trap ===\n");
    uint64 addr = r_csr_badv();
     struct VMA* vp = 0;
     //to fina the target vma
     for( int i=0 ; i<VMA_MAX ; i++ ) {
       if( p->vma[i].addr <= addr && addr < p->vma[i].addr + p->vma[i].len && p->vma[i].valid == 1 )
       {
         vp = &p->vma[i];
         break;
       }
     }
     if( vp != 0 )
     {
       uint64 mem = (uint64)kalloc();
       memset( (void*)mem , 0 , PGSIZE );
       if( -1 ==  mappages( p->pagetable , addr,  PGSIZE , mem , PTE_U | PTE_V | ( vp->prot << 1 ) ) )
         panic("pagefault map error");
       
       vp->mapcnt += PGSIZE; //maintain the mapcnt
       ilock( vp->f->ip );
       readi( vp->f->ip , 1 , addr , addr-vp->addr , PGSIZE); //copy a page of the file from the disk
       iunlock( vp->f->ip );
     }
  }
  else {
    printf("usertrap(): unexpected trapcause %p pid=%d\n", r_csr_estat(), p->pid);
    printf("            era=%p badi=%p\n", r_csr_era(), r_csr_badi());
    p->killed = 1;
  }

  if(p->killed)
    exit(-1);

  // give up the CPU if this is a timer interrupt.
  if(which_dev == 2)
  {
    myproc()->user_time++;
    yield();
  }

  usertrapret();
}

//
// return to user space
//
void
usertrapret(void)
{
  struct proc *p = myproc();

  // we're about to switch the destination of traps from
  // kerneltrap() to usertrap(), so turn off interrupts until
  // we're back in user space, where usertrap() is correct.
  intr_off();

  // set up trapframe values that uservec will need when
  // the process next re-enters the kernel.
  p->trapframe->kernel_pgdl = r_csr_pgdl();         // kernel page table
  p->trapframe->kernel_sp = p->kstack + PGSIZE; // process's kernel stack
  p->trapframe->kernel_hartid = r_tp();         // hartid for cpuid()
  // set up the registers that trampoline.S's sret will use
  // to get to user space.
  
  // set S Previous Privilege mode to User.
  uint32 x = r_csr_prmd();
  x |= PRMD_PPLV; // clear SPP to 0 for user mode
  x |= PRMD_PIE; // enable interrupts in user mode
  w_csr_prmd(x);

  // set S Exception Program Counter to the saved user pc.
  w_csr_era(p->trapframe->era);

  // tell trampoline.S the user page table to switch to.
    volatile uint64 pgdl = (uint64)(p->pagetable);

  // jump to uservec.S at the top of memory, which 
  // switches to the user page table, restores user registers,
  // and switches to user mode with ertn.
  userret(TRAPFRAME, pgdl);
}

// interrupts and exceptions from kernel code go here via kernelvec,
// on whatever the current kernel stack is.
void 
kerneltrap()
{
  int which_dev = 0;

  uint64 era = r_csr_era();
  uint64 prmd = r_csr_prmd();
  if((prmd & PRMD_PPLV) != 0)
    panic("kerneltrap: not from privilege0");
  if(intr_get() != 0)
    panic("kerneltrap: interrupts enabled");

  if((which_dev = devintr()) == 0){
    printf("estat %p\n", r_csr_estat());
    printf("era=%p eentry=%p\n", r_csr_era(), r_csr_eentry());
    panic("kerneltrap");
  }

  // give up the CPU if this is a timer interrupt.
  if(which_dev == 2 && myproc() != 0 && myproc()->state == RUNNING){
    myproc()->kernel_time++;
    yield();
  }

  // the yield() may have caused some traps to occur,
  // so restore trap registers for use by kernelvec.S's sepc instruction.
  w_csr_era(era);
  w_csr_prmd(prmd);
}

void
clockintr()
{
  acquire(&tickslock);
  ticks++;
  wakeup(&ticks);
  release(&tickslock);
}
void 
machine_trap()
{
  panic("machine error");
}

// check if it's an external interrupt or software interrupt,
// and handle it.
// returns 2 if timer interrupt,
// 1 if other device,
// 0 if not recognized.
// int
// devintr()
// {
//   uint32 estat = r_csr_estat();
//   // uint32 ecfg = r_csr_ecfg();
//   if(estat & HWI_VEC){
//     // this is a hardware interrupt, via IOCR.

//     // irq indicates which device interrupted.
//     // uint64 irq = extioi_claim();
//     int irq = UART0_IRQ;

//     if(irq == UART0_IRQ){
//       uartintr();

//     // tell the apic the device is
//     // now allowed to interrupt again.

//     }else if(irq){
//       printf("unexpected interrupt irq=%d\n", irq);
 
//     }
//     return 1;
//   } else if(estat & TI_VEC){
//     // software interrupt from a machine-mode timer interrupt,
//     // forwarded by timervec in kernelvec.S.

//     if(cpuid() == 0){
//       clockintr();
//     }
    
//     // acknowledge the software interrupt by clearing
//     // the SSIP bit in sip.
//     w_csr_ticlr(r_csr_ticlr() | CSR_TICLR_CLR);
//     //uartinit()
//     return 2;
//   } else {
//     return 0;
//   }
// }

int
devintr()
{
  uint32 estat = r_csr_estat();
  uint32 ecfg = r_csr_ecfg();


  if(estat & HWI_VEC & ecfg){
    // this is a supervisor external interrupt, via PLIC.
    // irq indicates which device interrupted.
    uint64 irq = extioi_claim();
    if(irq & (1UL << UART0_IRQ)){
      uartintr();

    // tell the apic the device is
    // now allowed to interrupt again.

      extioi_complete(1UL << UART0_IRQ);
    } else if(irq){
       printf("unexpected interrupt irq=%d\n", irq);

      apic_complete(irq); 
      extioi_complete(irq);        
    }
    return 1;
  } else if(estat & ecfg & TI_VEC){
    // software interrupt from a machine-mode timer interrupt,
    // forwarded by timervec in kernelvec.S.
    proc_tick();
    if(cpuid() == 0){
      clockintr();
    }
    w_csr_ticlr(r_csr_ticlr() | CSR_TICLR_CLR);
    return 2;
  } else {
    return 0;
  }
}
